/*

Copyright 2015-2021 Igor Petrovic

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

*/

#pragma once

#ifndef ANALOG_SUPPORTED
#include "Stub.h"
#else

#include <functional>
#include "database/Database.h"
#include "midi/src/MIDI.h"
#include "io/leds/LEDs.h"
#include "io/display/Display.h"
#include "io/common/CInfo.h"

namespace IO
{
    class Analog
    {
        public:
        enum class type_t : uint8_t
        {
            potentiometerControlChange,
            potentiometerNote,
            fsr,
            button,
            nrpn7b,
            nrpn14b,
            pitchBend,
            cc14bit,
            AMOUNT
        };

        enum class pressureType_t : uint8_t
        {
            velocity,
            aftertouch
        };

        enum class adcType_t : uint16_t
        {
            adc10bit = 1023,
            adc12bit = 4095
        };

        class HWA
        {
            public:
            //should return true if the value has been refreshed, false otherwise
            virtual bool value(size_t index, uint16_t& value) = 0;
        };

        class Filter
        {
            public:
            virtual Analog::adcType_t adcType()                                                                              = 0;
            virtual bool              isFiltered(size_t index, Analog::type_t type, uint16_t value, uint16_t& filteredValue) = 0;
            virtual void              reset(size_t index)                                                                    = 0;
        };

        using buttonHandler_t = std::function<void(size_t index, bool state)>;

        Analog(HWA&           hwa,
               Filter&        filter,
               Database&      database,
               MIDI&          midi,
               IO::LEDs&      leds,
               Display&       display,
               ComponentInfo& cInfo)
            : _hwa(hwa)
            , _filter(filter)
            , _database(database)
            , _midi(midi)
            , _leds(leds)
            , _display(display)
            , _cInfo(cInfo)
        {}

        void      update(bool forceResend = false);
        void      processReading(size_t index, uint16_t value);
        void      debounceReset(size_t index);
        void      registerButtonHandler(buttonHandler_t handler);
        adcType_t adcType();

        private:
        typedef struct
        {
            type_t   type;
            uint16_t lowerLimit;
            uint16_t upperLimit;
            uint16_t midiID;
            uint8_t  channel;
            bool     inverted;
        } analogDescriptor_t;

        void fillAnalogDescriptor(size_t index, analogDescriptor_t& descriptor);
        bool checkPotentiometerValue(size_t index, analogDescriptor_t& descriptor, uint16_t& value);
        bool checkFSRvalue(size_t index, analogDescriptor_t& descriptor, uint16_t& value);
        void sendMessage(size_t index, analogDescriptor_t& descriptor, uint16_t value);

        HWA&           _hwa;
        Filter&        _filter;
        Database&      _database;
        MIDI&          _midi;
        IO::LEDs&      _leds;
        Display&       _display;
        ComponentInfo& _cInfo;

        buttonHandler_t _buttonHandler                                                          = nullptr;
        uint8_t         _fsrPressed[MAX_NUMBER_OF_ANALOG]                                       = {};
        uint16_t        _lastValue[MAX_NUMBER_OF_ANALOG + MAX_NUMBER_OF_TOUCHSCREEN_COMPONENTS] = {};
    };
}    // namespace IO

#endif